#! /usr/bin/env python

from os.path import dirname, basename, expanduser
import os
import errno
from argparse import ArgumentParser
from getpass import getpass

import yaml
import requests
from bs4 import BeautifulSoup

def load_config():
    try:
        with open("config.yml") as f:
            config = yaml.safe_load(f)
    except IOError:
        xdg_config_home = os.getenv("XDG_CONFIG_HOME", expanduser("~/.config"))
        config_path = xdg_config_home + "/gog/config.yml"
        
        with open(config_path) as f:
            config = yaml.safe_load(f)
    
    return config

def wget(dryrun, output, url):
    cmd = "wget -c -O %s '%s'" % (output, url)
    if dryrun:
        print cmd
    else:
        try:
            os.makedirs(dirname(output))
        except OSError as e:
            if not e.errno == errno.EEXIST:
                raise
        
        os.system(cmd)
    
class Filter(object):
    """ Filter object
    """
    
    def __init__(self, filter):
        """ Create a filter given in a string
        
        Arguments:
        - `filter`: string from the config file
        """
        self._filter = filter
        
        self._included = []
        self._excluded = []
        if filter:
            for p in filter.split(" "):
                if p[0] == "-":
                    self._excluded.append(p[1:])
                else:
                    self._included.append(p)
    
    def __contains__(self, item):
        if item is None:
            return True
        
        if item in self._excluded:
            return False
        if len(self._included) > 0 and item not in self._included:
            return False
        
        return True
    
    def __str__(self):
        return "Filter(%s)" % self._filter

class Item(object):
    """ Item to download
    """
    
    def __init__(self, gog, element, platform):
        """ Build an element from an html element
        
        Arguments:
        - `gog`: GOG object
        - `element`: BeatufilSoup element
        """
        self._gog     = gog
        self._element = element
        
        self.url      = element.get("href")
        self.name     = element.find("span", "details-underline").string
        self.size     = element.find("span", "size").string
        self.platform = platform
        
        req = self._gog._session.head(self.url)
        assert(req.status_code == 302)
        
        self.filename = basename(req.headers["location"])
        self.filename = self.filename[:self.filename.find("?")]
        
        if element.get("class")[0] == "list_game_item":
            self.lang = element.find_parent("div").get("class")[1][5:] # remove lang_ part
        else:
            self.lang = None
        
        if element.get("class")[0] == "list_bonus_item":
            self.type = element.get("class")[1][1:]
        else:
            self.type = None
    
    def __str__(self):
        return "  - %s (size: %s, language: %s, type: %s, platform: %s, filename: %s)" % (self.name, self.size, self.lang, self.type, self.platform, self.filename)
    
    def download_url(self):
        req = self._gog._session.head(self.url)
        assert(req.status_code == 302)
        return req.headers["location"]

class Game(object):
    """ Game representation
    """
    _base_url   = "http://gog.com"
    
    def __init__(self, gog, element):
        """ Build a game representation with an html element
        
        Arguments:
        - `gog`: GOG object
        - `element`: BeautifulSoup element
        """
        self._gog       = gog
        self._session   = gog._session
        self._element   = element
        
        self.title      = element.find("span", "game-title-link").string
        self.thumb      = self._base_url + element.find("img", "list_image").get("src")
        self.background = self._base_url + element.get("data-background")
        self.tags       = element.get("data-title")
        self.id         = element.get("id")[8:] # remove the game_li_ part
    
    def __str__(self):
        self.details()
        return "- %s (%s)\n  %s\n  %s\n  %s\n  %s" % (self.title.encode("UTF-8"), self.name, self.tags, self.url, self.thumb, self.background)
    
    def details(self):
        url = gog._base_url + "/account/ajax?a=gamesListDetails&g=%s" % self.id
        req = self._session.get(url)
        assert(req.status_code == 200)
        soup = BeautifulSoup(req.json()["details"]["html"])
        
        self.purchase_date  = soup.find("div", "list_purchased").string
        self.url            = soup.find("h2").find("a").get("href")
        self.name           = basename(self.url)
        
        list_down_browser   = soup.find("div", "list_down_browser")
        win_download        = list_down_browser.find("div", "win-download")
        win_installers      = [Item(self._gog, item, "windows") for item in win_download.find_all("a", "list_game_item")]
        mac_download        = list_down_browser.find("div", "mac-download")
        mac_installers      = [Item(self._gog, item, "mac") for item in mac_download.find_all("a", "list_game_item")]
        self.installers     = win_installers + mac_installers
        
        bonus_content_list  = soup.find("div", "bonus_content_list browser")
        if bonus_content_list is None:
            self.bonuses    = []
        else:
            self.bonuses    = [Item(self._gog, item, None) for item in bonus_content_list.find_all("a", "list_bonus_item")]
    
class GOG(object):
    """ GOG user account
    """
    _base_url   = "https://secure.gog.com"
    
    def __init__(self, config):
        """ Create the GOG object
        
        Arguments:
        - `config`: configuration from the configuration file
        """
        
        if "filters" not in config:
            config["filters"] = {}
        
        self._filters = {}
        for field in ["games", "bonus", "platforms", "languages"]:
            if field not in config["filters"]:
                config["filters"][field] = None
            
            self._filters[field] = Filter(config["filters"][field])
        
        self._directory = expanduser(config["directory"])
        self._dryrun    = config["dryrun"]
        
        self._session = requests.Session()
        
        self._find_gutm()
        self._find_buk()
    
    def _find_gutm(self):
        req = self._session.get(self._base_url)
        assert(req.status_code == 200)
        page = BeautifulSoup(req.content)
        self._gutm = page.find("input", id="gutm").get("value")
    
    def _find_buk(self):
        # from
        # http://static.gog.com/www/default/-js/_min_global.ab1acd31.js
        # or
        # http://static.gog.com/www/default/-js/global.ab1acd31.js
        
        url = self._base_url + "/user/ajax"
        payload = {
            "a"     : "get",
            "c"     : "frontpage", #__ctr,
            "p1"    : False,
            "p2"    : False,
            "auth"  : "",
            "games" : "", #_pageGamesList.join(", "),
            "gutm"  : self._gutm,
            "pp"    : ""
        }
        req = self._session.post(url, data=payload)
        assert(req.status_code == 200)
        self._buk = req.json()["buk"]
    
    def login(self, email, password):
        url = self._base_url + "/login"
        payload = {
            "log_email"      : email,
            "log_password"   : password,
            "buk"            : self._buk,
            "redirectOk"     : "/",
            "unlockSettings" : "1",
        }
        req = self._session.post(url, data=payload)
    
    def games(self):
        url = self._base_url + "/account"
        req = self._session.get(url)
        assert(req.status_code == 200)
        page = BeautifulSoup(req.content)
        gamesList = page.find("div", id="gamesList")
        
        for g in gamesList.find_all("div", "game-item"):
            game = Game(self, g)
            game.details()
            if game.name in self._filters["games"]:
                yield game
    
    def filter(self, type):
        return self._filters[type]
    
    def show(self):
        for g in gog.games():
            print g
            for item in g.installers:
                if item.platform in gog.filter("platforms") and item.type in gog.filter("bonus") and item.lang in gog.filter("languages"):
                    print item
            for item in g.bonuses:
                if item.platform in gog.filter("platforms") and item.type in gog.filter("bonus") and item.lang in gog.filter("languages"):
                    print item
    
    def mirror(self):
        cookies = None
        
        try:
            os.makedirs(self._directory)
        except OSError as e:
            if not e.errno == errno.EEXIST:
                raise
        os.chdir(self._directory)
        
        for g in gog.games():
            for item in g.installers:
                if item.platform in gog.filter("platforms") and item.type in gog.filter("bonus") and item.lang in gog.filter("languages"):
                    filename = g.name + "/" + item.filename
                    url = item.download_url()
                    wget(self._dryrun, filename, url)
            
            for item in g.bonuses:
                if item.platform in gog.filter("platforms") and item.type in gog.filter("bonus") and item.lang in gog.filter("languages"):
                    filename = g.name + "/" + item.filename
                    url = item.download_url()
                    wget(self._dryrun, filename, url)

if __name__ == "__main__":
    parser = ArgumentParser(description='Heavy client for gog.com')
    parser.add_argument('--show', action="store_true",
                        help='show the content of your library')
    parser.add_argument('--mirror', action="store_true",
                        help='mirror your library')
    parser.add_argument('--all', action="store_true",
                        help='disable all filters')
    parser.add_argument('--all-games', dest="all_games", action="store_true",
                        help='disable all filters on games')
    parser.add_argument('--dry-run', dest="dryrun", action="store_true",
                        help='just prints the action, but does nothing')
    parser.add_argument('--config',
                        help='configuration file')
    args = parser.parse_args()
    
    config = load_config()
    
    config["dryrun"] = args.dryrun
    
    if "filters" not in config:
        config["filters"] = {}
    if args.all:
        config["filters"] = {}
    if args.all_games:
        config["filters"]["games"] = {}
    
    if args.show or args.mirror:
        if "user" not in config:
            config["user"] = {}
        if "email" not in config["user"]:
            config["user"]["email"] = raw_input("E-mail: ")
        if "password" not in config["user"]:
            config["user"]["password"] = getpass()
        
        gog = GOG(config)
        gog.login(config["user"]["email"], config["user"]["password"])
    
    if args.show:
        gog.show()
    elif args.mirror:
        gog.mirror()
    else:
        print "No action provided"

